package com.all.bukkit.threadcheck;

import java.util.List;

import org.bukkit.craftbukkit.scheduler.CraftWorker;
import org.bukkit.craftbukkit.util.ServerShutdownThread;
import org.bukkit.event.Event;
import org.bukkit.event.Event.Priority;
import org.bukkit.event.Listener;
import org.bukkit.plugin.java.JavaPlugin;

import com.google.common.collect.Lists;

public class ThreadCheck {

	@SuppressWarnings("unused")
	private static ThreadCheck checker;
	
	public static void init(JavaPlugin plugin) {
		try {
			checker = new ThreadCheck(plugin);
		} catch (Throwable t) {
			System.out.println("Failed to initialize thread checker:");
			t.printStackTrace();
		}
	}
	public static void deinit() {
		checker = null;
	}
		
	private ThreadCheck(JavaPlugin p) {
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_ITEM_HELD);
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_JOIN);
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_QUIT);
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_KICK);
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_TELEPORT);
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_MOVE);
	    this.registerEvent(p, playerListener, Event.Type.PLAYER_CHANGED_WORLD);
	    this.registerEvent(p, worldListener, Event.Type.CHUNK_LOAD);
	    this.registerEvent(p, worldListener, Event.Type.CHUNK_UNLOAD);
	    this.registerEvent(p, worldListener, Event.Type.CHUNK_POPULATED);
	    this.registerEvent(p, worldListener, Event.Type.WORLD_LOAD);
	    this.registerEvent(p, worldListener, Event.Type.WORLD_UNLOAD);
	    this.registerEvent(p, worldListener, Event.Type.WORLD_INIT);
	    this.registerEvent(p, worldListener, Event.Type.WORLD_SAVE);
	    this.registerEvent(p, worldListener, Event.Type.SPAWN_CHANGE);
	    this.registerEvent(p, entityListener, Event.Type.CREATURE_SPAWN);
	    this.registerEvent(p, entityListener, Event.Type.ITEM_SPAWN);
	    this.registerEvent(p, entityListener, Event.Type.PAINTING_BREAK);
	    this.registerEvent(p, entityListener, Event.Type.PAINTING_PLACE);
	    this.registerEvent(p, vehicleListener, Event.Type.VEHICLE_MOVE);
	    this.registerEvent(p, vehicleListener, Event.Type.VEHICLE_CREATE);
	    this.registerEvent(p, vehicleListener, Event.Type.VEHICLE_DESTROY);
	    this.registerEvent(p, vehicleListener, Event.Type.VEHICLE_ENTER);
	    this.registerEvent(p, vehicleListener, Event.Type.VEHICLE_EXIT);
	}
	
	public static void setAllowedMethods(List<String> crossThreadExceptions) {
		allowedMethods = new String[crossThreadExceptions.size()];
		for (int i = 0; i < allowedMethods.length; i++) {
			allowedMethods[i] = crossThreadExceptions.get(i).toLowerCase();
		}
	}
	
	private void registerEvent(JavaPlugin p, Listener listener, Event.Type type) {
		p.getServer().getPluginManager().registerEvent(type, listener, Priority.Lowest, p);
	}
	
	private final PlayerListener playerListener = new PlayerListener();
	private final WorldListener worldListener = new WorldListener();
	private final EntityListener entityListener = new EntityListener();
	private final VehicleListener vehicleListener = new VehicleListener();

	private static final Thread currentThread = Thread.currentThread();

	private static StackTraceElement[] previouserror;
	private static String[] allowedMethods;
	public static void check() {
        check(false);
	}
	
	private static String getListenerName(String classname) {
		if (classname.equals("PlayerListener")) return "org.bukkit.event.player.PlayerListener";
		if (classname.equals("EntityListener")) return "org.bukkit.event.entity.EntityListener";
		if (classname.equals("VehicleListener")) return "org.bukkit.event.vehicle.VehicleListener";
		if (classname.equals("WorldListener")) return "org.bukkit.event.world.WorldListener";
		if (classname.equals("WeatherListener")) return "org.bukkit.event.weather.WeatherListener";
		return "org.bukkit.event." + classname;
	}
	
    //java.lang.Thread.run(Unknown Source)
    //org.bukkit.craftbukkit.scheduler.CraftWorker.run(CraftWorker.java:34)
	@SuppressWarnings("rawtypes")
	private static boolean classCheck(StackTraceElement element, Class classtocheck) {
		return element.getClassName().equals(classtocheck.getName());
	}
	public static void check(boolean isEvent) {
		final Thread t = Thread.currentThread();
		if (t != currentThread) {
			List<StackTraceElement> stack = Lists.newArrayList(t.getStackTrace());
			stack.remove(0); //remove check function from stacktrace
			stack.remove(0); //remove check function from stacktrace
			if (t.getClass().equals(ServerShutdownThread.class)) return;
			String classname;
			//general thread?
			if (classCheck(stack.get(stack.size() - 1), Thread.class)) {
				stack.remove(stack.size() - 1);
				//bukkit task?
				if (classCheck(stack.get(stack.size() - 1), CraftWorker.class)) {
					classname = stack.get(stack.size() - 2).getClassName();
				} else {
					classname = stack.get(stack.size() - 1).getClassName();
				}
			} else {
				classname = t.getClass().getName();
			}
			if (isEvent) {
				//replace call
				final StackTraceElement event = stack.get(0);
				final String mname = event.getMethodName();
				String cname = event.getClassName();
				cname = cname.substring(cname.lastIndexOf('.') + 1);
				stack.set(0, new StackTraceElement(getListenerName(cname), mname, cname, 0));
			}

			//allowed?
			final boolean checkall = allowedMethods != null && allowedMethods.length != 0;;
			final boolean checkprev = previouserror != null && previouserror.length == stack.size();
			if (checkall || checkprev) {
				//equal?
				boolean eq = checkprev;
				for (int i = 0; i < stack.size(); i++) {
					StackTraceElement el = stack.get(i);
					if (checkall) {
						for (String e : allowedMethods) {
							if (el.toString().toLowerCase().contains(e)) return;
						}
					}
					if (eq && !previouserror[i].equals(el)) {
						eq = false;
					}
				}
				if (eq) return;
			}
			if (!checkprev) previouserror = new StackTraceElement[stack.size()];
			for (int i = 0; i < stack.size(); i++) previouserror[i] = stack.get(i);
			IllegalAccessError err = new IllegalAccessError("Synchronized code got accessed from another thread: " + classname);
			err.setStackTrace(previouserror);
			throw err;
		}
	}
}
