package com.bergerkiller.bukkit.nolagg;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.Player;

import org.bukkit.craftbukkit.entity.CraftPlayer;

import com.all.bukkit.threadcheck.ThreadCheck;

import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkCoordIntPair;
import net.minecraft.server.EntityPlayer;
import net.minecraft.server.NetworkManager;
import net.minecraft.server.Packet53BlockChange;
import net.minecraft.server.World;

@SuppressWarnings({"rawtypes", "unchecked"})
public class ChunkSendQueue extends LinkedList {

	private static ArrayList<Operation> syncedops;
	
	private static Field queuesizefield;
	public static boolean enabled = true;
	private static final long serialVersionUID = 1L;
	private static int task = -1;
	public static void init() {
		syncedops = new ArrayList<Operation>();
		try {
			queuesizefield = NetworkManager.class.getDeclaredField("x");
			queuesizefield.setAccessible(true);
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to hook into the player packet queue size field");
			NoLagg.log(Level.SEVERE, "Distortions in the chunk rate will cause players to get kicked");
		}
		task = Bukkit.getServer().getScheduler().scheduleSyncRepeatingTask(NoLagg.plugin, new Runnable() {
			public void run() {
				try {
					//synced ops
					synchronized (syncedops) {
						for (Operation op : syncedops) op.run();
						syncedops.clear();
					}
					if (enabled) {
						global(new Operation() {
							public void handle(EntityPlayer ep) {
								bind(ep).update();
							}
						});
					}
				} catch (Exception ex) {
					NoLagg.log(Level.SEVERE, "An error occured while sending chunks:");
					ex.printStackTrace();
				}
			}
		}, 1, 1);
	}
	public static void deinit() {
		if (task != -1) {
			Bukkit.getServer().getScheduler().cancelTask(task);
		}
		syncedops = null;
	}
	
	public static void scheduleQueueGlobal(final Chunk chunk) {
		scheduleQueueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void scheduleQueueGlobal(final World world, final int x, final int z) {
		synchronized(syncedops) {
			syncedops.add(new Operation() {
				public void init() {
					queueGlobal(world, x, z);
				}
			});
		}
	}
	
	public static void deQueueGlobal(final Chunk chunk) {
		deQueueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void queueGlobal(final Chunk chunk) {
		queueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void deQueueGlobal(final World world, final int x, final int z) {
		deQueueGlobal(world, new ChunkCoordIntPair(x, z));
	}
	public static void queueGlobal(final World world, final int x, final int z) {
		queueGlobal(world, new ChunkCoordIntPair(x, z));
	}
	public static void deQueueGlobal(final World world, final ChunkCoordIntPair pair) {
		ThreadCheck.check();
		world(world, new Operation() {
			public void handle(EntityPlayer ep) {
				ep.chunkCoordIntPairQueue.remove(pair);
			}
		});
	}
	public static void queueGlobal(final World world, final ChunkCoordIntPair pair) {
		ThreadCheck.check();
		world(world, new Operation() {
			public void handle(EntityPlayer ep) {
		        if (Math.abs(pair.x - ((int) ep.locX >> 4)) > Util.view) return;
		        if (Math.abs(pair.z - ((int) ep.locZ >> 4)) > Util.view) return;
				ep.chunkCoordIntPairQueue.add(0, pair);
			}
		});
	}
	
	private ValueWatcher queueSizeWatcher = new ValueWatcher(5000, 100, 50000, 200000);
	private int getQueueSize() {
		if (queuesizefield != null) {
			try {
				return (Integer) queuesizefield.get(this.ep.netServerHandler.networkManager);
			} catch (Throwable t) {
				queuesizefield = null;
				NoLagg.log(Level.SEVERE, "Failed to hook into the player packet queue size field");
				NoLagg.log(Level.SEVERE, "Distortions in the chunk rate will cause players to get kicked");
			}
		}
		return 0;
	}
	
	private static <T extends Operation> T global(T op) {
		op.execute();
		return op;
	}
	private static <T extends Operation> T world(World world, T op) {
		op.execute(world);
		return op;
	}
	private static class Operation {
		public void handle(EntityPlayer ep) {}
		public void init(World world) {}
		public void init() {}
		public void execute(World world) {
			ThreadCheck.check();
			this.init(world);
			try {
				for (Object o : world.players) this.handle((EntityPlayer) o);
			} catch (ConcurrentModificationException ex) {
				NoLagg.log(Level.SEVERE, "A plugin tried to modify 'World.players' (world entity list) from an async thread!");
			}
		}
		public void execute() {
			ThreadCheck.check();
			this.init();
			for (World world : Util.getWorlds()) {
				this.execute(world);
			}
		}
		public Operation run() {
			this.execute();
			return this;
		}
	}
	
	public static ChunkSendQueue bind(Player with) {
		if (!enabled) return null;
		return bind(((CraftPlayer) with).getHandle());
	}
	public static ChunkSendQueue bind(EntityPlayer with) {
		if (!enabled) return null;
		if (!(with.chunkCoordIntPairQueue instanceof ChunkSendQueue)) {
			ThreadCheck.check();
			with.chunkCoordIntPairQueue = new ChunkSendQueue(with);
		}
		return (ChunkSendQueue) with.chunkCoordIntPairQueue;
	}	
	private ChunkSendQueue(EntityPlayer ep) {
		this.ep = ep;
	}
	
	public static int initialBoost = 0; //the size that has to be downloaded
	public static double maxRate = 2;
	public static double minRate = 0.25;
			
	public static double getAverageRate() {
		return global(new Operation() {
			double totalrate = 0;
			int pcount = 0;
			public void handle(EntityPlayer ep) {
				totalrate += bind(ep).rate;
				pcount++;
			}
			public double get() {
				return this.totalrate / (double) pcount;
			}
		}).get();
	}
	public double getRate() {
		return this.rate;
	}
	
	private EntityPlayer ep;
	private BlockFace sendDirection = BlockFace.NORTH;
	private int x;
	private int z;
	private int boost = 0;
	private double rate = (minRate + maxRate) / 2;
	
	/*
	 * Used to send packets and triggers
	 */
	private List<ChunkCoordIntPair> toTrigger = new ArrayList<ChunkCoordIntPair>();
	private int intervalcounter = -1;
	private void update() {
		this.queueSizeWatcher.update(this.getQueueSize());
		if (this.queueSizeWatcher.isMaxed()) {
			this.rate = minRate;
		} else if (this.queueSizeWatcher.isMinimized()) {
			this.rate = maxRate;
		} else {
			//a simple trigger to get some movement in here...
			if (this.queueSizeWatcher.getValue() > 100000) {
				this.rate -= 0.01;
			} else if (this.queueSizeWatcher.getValue() < 10000) {
				this.rate += 0.01;
			}
			//adjustment of the rate
			this.rate -= this.queueSizeWatcher.getRate();
			//set rate bounds
			if (this.rate > maxRate) {
				this.rate = maxRate;
			} else if (this.rate < minRate) {
				this.rate = minRate;
			}
		}

		//send chunks
		if (this.rate >= 1) {
			this.update(1, (int) this.rate);
		} else {
			this.update((int) (1 / this.rate), 1);
		}
	}
	public Object[] toArray(Object[] value) {
		ThreadCheck.check();
		return super.toArray(value);
	}
	private void update(int interval, int rate) {
		if (interval == 0) interval = 1;
		if (rate == 0) rate = 1;
		if (this.intervalcounter >= interval)  {
			//sorting
			BlockFace newDirection = Util.getDirection((int) this.ep.yaw - 90);
			int newx = (int) (ep.locX + ep.motX * 10) >> 4;
			int newz = (int) (ep.locZ + ep.motZ * 10) >> 4;
			if (newx != this.x || newz != this.z || this.sendDirection != newDirection) {
				this.sendDirection = newDirection;
				this.x = newx;
				this.z = newz;
				//sort self
				if (!this.isEmpty()) {
					try {
						java.util.Collections.sort(this, new ChunkCoordComparator(this.x, this.z, this.sendDirection));
					} catch (ConcurrentModificationException ex) {
						NoLagg.log(Level.SEVERE, "A plugin tried to modify 'EntityPlayer.chunkCoordIntPairQueue' from another thread!");
					} catch (ArrayIndexOutOfBoundsException ex) {
						NoLagg.log(Level.SEVERE, "A plugin tried to modify 'EntityPlayer.chunkCoordIntPairQueue' from another thread!");
					}
					//download speed boost when the chunk below the player has to be sent
					this.updateBoost();
				}
			}

			//send triggers
			for (ChunkCoordIntPair pair : this.toTrigger) {
				//send a single block packet for this chunk
				this.ep.netServerHandler.sendPacket(new Packet53BlockChange(pair.x * 16, 0, pair.z * 16, this.ep.world));
			}
			this.toTrigger.clear();

			//send using rate
			while (!this.isEmpty()) {
				if (this.boost > 0) {
					--this.boost;
				} else if (rate > 0) {
					--rate;
				} else {
					break;
				}
				this.sendFirst();
			}
			this.intervalcounter = 1;
		} else {
			this.intervalcounter++;
		}
	}
	private void sendFirst() {
		if (this.ep == null || this.ep.world == null) return;
		ChunkCoordIntPair pair = (ChunkCoordIntPair) super.remove(0);
		if (pair == null) return;
		ChunkHandler.sendChunk(this.ep.world.getChunkAt(pair.x, pair.z), ep);
		this.toTrigger.add(pair);
	}
	private void updateBoost() {
		if (this.isEmpty()) return;
		ChunkCoordIntPair first = (ChunkCoordIntPair) super.get(0);
		if (first.x == this.x && first.z == this.z) this.boost = initialBoost;
	}
		
	public boolean containsAll(Collection coll) {
		ThreadCheck.check();
		return super.containsAll(coll);
	}
	public boolean contains(Object o) {
		ThreadCheck.check();
		return super.contains(o);
	}
	public boolean addAll(Collection coll) {
		ThreadCheck.check();
		return super.addAll(coll);
	}
	public boolean addAll(int index, Collection coll) {
		ThreadCheck.check();
		return super.addAll(index, coll);
	}
	
	public void clear() {
		ThreadCheck.check();
		super.clear();
	}
	
	/*
	 * get(0) is called in EntityPlayer to get the next chunk to send
	 */
	public Object get(int index) {
		ThreadCheck.check();
		this.updateBoost();
		if (this.boost > 0) {
			--this.boost;
			this.sendFirst();
		}
		return null;
	}
	
	public boolean remove(Object object) {
		ThreadCheck.check();
		return super.remove(object);
	}
	
	/*
	 * add(0) is called in PlayerInstance to queue a new chunk coordinate
	 */
	public boolean add(Object object) {
		ThreadCheck.check();
		if (object == null) return false;
		ChunkCoordIntPair pair = (ChunkCoordIntPair) object;
        if (Math.abs(pair.x - ((int) this.ep.locX >> 4)) > Util.view) return false;
        if (Math.abs(pair.z - ((int) this.ep.locZ >> 4)) > Util.view) return false;
		this.sendDirection = null; //invalidate
		return super.add(object);
	}

}
